/* monkey.c
 *
 * The FTP Monkey can be used to simulate a FTP user.  It opens a host,
 * randomly changes directories and fetches files.
 */

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <ncftp.h>				/* Library header. */
#include <Strn.h>				/* Library header. */
#include <sio.h>				/* Library header. */

#if defined(WIN32) || defined(_WINDOWS)
#	define sleep WinSleep
#endif

#define kYesCloseYesQuit 0
#define kYesCloseNoQuit 1
#define kNoCloseNoQuit 2

LineList gDir;
int gHaveDir = 0;
int gQuitMode = kYesCloseYesQuit;
int gPrintLists = 0;
int gLoops = 0;
int gGotSig = 0;
int gMaxLoops = 0;
int gCloseAndReconnect = 0;
int gNoMlst = 1;
time_t gAlarmClock = 0;
const char *gMainDir = "/pub";
char gPasswd[64];
FTPLibraryInfo li;
FTPConnectionInfo fi;

#define kErrStream stdout

#define c_quit 0
#define c_cd 1
#define c_pwd 2
#define c_dir 3
#define c_get 4
#define c_max 5

int gWeights[c_max] = {
	10,	/* quit */
	20,	/* cd */
	5,	/* pwd */
	15,	/* dir */
	50,	/* get */
};

static void
Usage(void)
{
	fprintf(kErrStream, "FTP Monkey, copyright 1996 by Mike Gleason, NCEMRSoft.\n");
	fprintf(kErrStream, "Usage: ftp_monkey [flags] hostname\n");

	(void) fprintf(kErrStream, "\nGeneral Flags:\n\
  -u XX  Use username XX instead of anonymous.\n\
  -p XX  Use password XX with the username.\n\
  -P XX  Use port number XX instead of the default FTP service port (21).\n\
  -d XX  Use the file XX for debug logging.\n");
	(void) fprintf(kErrStream, "\nMonkey Flags:\n\
  -a XX  Quit after test has run for XX seconds.\n\
  -m XX  Quit after test has run for XX operations.\n\
  -s XX  Set random seed to XX.\n\
  -D XX  Use remote directory XX as base.\n\
  -l     Print the output from directory listings.\n\
  -Q     Monkey may not end FTP session nor quit.\n\
  -R     Monkey may end FTP session, but must reconnect and start a new one.\n\
  -S     Monkey may end FTP session and quit (default).\n");
	(void) fprintf(kErrStream, "\nNotes:\n\
  The default behavior (-S) is to have Monkey simulate exactly one FTP session\n\
  and quit after a random number of operations.  To have Monkey simulate exactly\n\
  one FTP session and stay logged in forever (or until the limits specified by\n\
  -a or -m are met), use -Q mode.  To have Monkey continually simulate FTP\n\
  sessions of random number of operations, use -R.\n\n");
	fprintf(kErrStream, "Library version: %s.\n", gLibNcFTPVersion + 5);
#ifdef UNAME
	fprintf(kErrStream, "System: %s.\n", UNAME);
#endif
	DisposeWinsock();
	exit(2);
}	/* Usage */




static char *
GetPass2(const char *const prompt)
{
#ifdef HAVE_GETPASS
	return getpass(prompt);
#elif defined(_CONSOLE) && (defined(WIN32) || defined(_WINDOWS))
	static char pwbuf[128];
	char *dst, *dlim;
	int c;

	(void) memset(pwbuf, 0, sizeof(pwbuf));
	if (! _isatty(_fileno(stdout)))
		return (pwbuf);
	(void) fputs(prompt, stdout);
	(void) fflush(stdout);

	for (dst = pwbuf, dlim = dst + sizeof(pwbuf) - 1;;) {
		c = _getch();
		if ((c == 0) || (c == 0xe0)) {
			/* The key is a function or arrow key; read and discard. */
			(void) _getch();
		}
		if ((c == '\r') || (c == '\n'))
			break;
		if (dst < dlim)
			*dst++ = (char) c;
	}
	*dst = '\0';

	(void) fflush(stdout);
	(void) fflush(stdin);
	return (pwbuf);
#else
	static char pwbuf[128];

	(void) memset(pwbuf, 0, sizeof(pwbuf));
#if defined(WIN32) || defined(_WINDOWS)
	if (! _isatty(_fileno(stdout)))
		return (pwbuf);
#else
	if (! isatty(STDERR_FILENO))
		return (pwbuf);
#endif
	(void) fputs(prompt, stdout);
	(void) fflush(stdout);
	(void) FGets(pwbuf, sizeof(pwbuf), stdin);
	(void) fflush(stdout);
	(void) fflush(stdin);
	return (pwbuf);
#endif
}	/* GetPass2 */



static int
Ri(int a)
{
	int b;

	b = rand() % a;
	return (b);
}	/* Ri */


static int
Rp(int a)
{
	int b;

	b = rand() % 100;
	if (b < a)
		return (1);
	return (0);
}	/* Rp */



static int
FTPReadOneFile(const FTPCIPtr cip, const char *file)
{
	char *buf;
	size_t bufSize;
	int tmpResult, result;
	int nread;

	if (cip == NULL)
		return (kErrBadParameter);
	if (strcmp(cip->magic, kLibraryMagic))
		return (kErrBadMagic);
	
	if (file == NULL)
		return (kErrBadParameter);
	if (file[0] == '\0')
		return (kErrBadParameter);

	result = kNoErr;
	FTPInitIOTimer(cip);
	FTPStartIOTimer(cip);
	tmpResult = FTPStartDataCmd(cip, kNetReading, kTypeBinary, (longest_int) 0, "RETR %s", file);
	if (tmpResult < 0) {
		result = tmpResult;
		if (result == kErrGeneric)
			result = kErrRETRFailed;
		if ((cip->errNo == kErrAcceptDataSocket) || (cip->errNo == kErrNewStreamSocket))
			FTPCloseHost(cip);
		cip->errNo = result;
		return (result);
	}

	buf = cip->buf;
	bufSize = cip->bufSize;
	for ( ; ; ) {
#ifdef NO_SIGNALS
		nread = SRead(cip->dataSocket, buf, bufSize, (int) cip->xferTimeout, kFullBufferNotRequired|kNoFirstSelect);
		if (nread == kTimeoutErr) {
			cip->errNo = result = kErrDataTimedOut;
			/* Error(cip, kDontPerror, "Remote write timed out.\n"); */
			break;
		} else if (nread < 0) {
			if (errno == EINTR)
				continue;
			/* Error(cip, kDoPerror, "Remote read failed.\n"); */
			result = kErrSocketReadFailed;
			cip->errNo = kErrSocketReadFailed;
			break;
		} else if (nread == 0) {
			break;
		}
#else
		nread = read(cip->dataSocket, buf, bufSize);
		if (nread < 0) {
			/* Error(cip, kDoPerror, "Remote read failed.\n"); */
			result = kErrSocketReadFailed;
			cip->errNo = result;
			break;
		} else if (nread == 0) {
			break;
		}
#endif	/* NO_SIGNALS */
		cip->bytesTransferred += (long) nread;
	}

	tmpResult = FTPEndDataCmd(cip, 1);
	if ((tmpResult < 0) && (result == 0)) {
		result = kErrRETRFailed;
		if ((cip->errNo == kErrAcceptDataSocket) || (cip->errNo == kErrNewStreamSocket))
			FTPCloseHost(cip);
		cip->errNo = result;
	}

	FTPStopIOTimer(cip);
	return (result);
}	/* FTPReadOneFile */



static int
FileLIST(const FTPCIPtr cip, LineListPtr lines, const char *lsflags)
{
	char secondaryBuf[512];
	char line[128];
	char *tok;
	int ftype;
	char fname[128];
#ifndef NO_SIGNALS
	char *secBufPtr, *secBufLimit;
	int nread;
	int result;
#else	/* NO_SIGNALS */
	SReadlineInfo lsSrl;
	int result;
#endif	/* NO_SIGNALS */

	InitLineList(lines);

	FTPInitIOTimer(cip);
	FTPStartIOTimer(cip);
	result = FTPStartDataCmd(cip, kNetReading, kTypeAscii, (longest_int) 0, "%s%s", "LIST", lsflags);

#ifdef NO_SIGNALS

	if (result == 0) {
		if (InitSReadlineInfo(&lsSrl, cip->dataSocket, secondaryBuf, sizeof(secondaryBuf), (int) cip->xferTimeout, 1) < 0) {
			/* Not really fdopen, but close in what we're trying to do. */
			result = kErrFdopenR;
			cip->errNo = kErrFdopenR;
			/* Error(cip, kDoPerror, "Could not fdopen.\n"); */
			return (result);
		}
		
		for (;;) {
skip:
			result = SReadline(&lsSrl, line, sizeof(line) - 2);
			if (result == kTimeoutErr) {
				/* timeout */
				/* Error(cip, kDontPerror, "Could not directory listing data -- timed out.\n"); */
				cip->errNo = kErrDataTimedOut;
				return (cip->errNo);
			} else if (result == 0) {
				/* end of listing -- done */
				cip->numListings++;
				break;
			} else if (result < 0) {
				/* error */
				/* Error(cip, kDoPerror, "Could not read directory listing data"); */
				result = kErrLISTFailed;
				cip->errNo = kErrLISTFailed;
				break;
			}

			if (line[result - 1] == '\n') {
				line[result - 1] = '\0';
			}

			cip->bytesTransferred += (long) result;
			ftype = line[0];
			if ((ftype == '-') || (ftype == 'd')) {
				tok = line + strlen(line);
				while (*--tok != ' ') {
					if (tok < line)
						goto skip;
				}
				fname[0] = (char) ftype;
				fname[1] = (char) ' ';
				fname[2] = '\0';
				STRNCAT(fname, tok + 1);
				AddLine(lines, fname);
			}
		}

		DisposeSReadlineInfo(&lsSrl);
		result = FTPEndDataCmd(cip, 1);
		FTPStopIOTimer(cip);
		if (result < 0) {
			result = kErrLISTFailed;
			cip->errNo = result;
		}
		result = kNoErr;
	} else if (result == kErrGeneric) {
		result = kErrLISTFailed;
		if ((cip->errNo == kErrAcceptDataSocket) || (cip->errNo == kErrNewStreamSocket))
			FTPCloseHost(cip);
		cip->errNo = result;
	}

#else	/* NO_SIGNALS */	
	
	if (result == 0) {
		/* This line sets the buffer pointer so that the first thing
		 * BufferGets will do is reset and fill the buffer using
		 * real I/O.
		 */
		secBufPtr = secondaryBuf + sizeof(secondaryBuf);
		secBufLimit = (char *) 0;

		for ( ; ; ) {
skip:
			nread = BufferGets(line, sizeof(line), cip->dataSocket, secondaryBuf, &secBufPtr, &secBufLimit, sizeof(secondaryBuf));
			if (nread <= 0) {
				if (nread < 0)
					break;
			} else {
				cip->bytesTransferred += (long) nread;
				ftype = line[0];
				if ((ftype == '-') || (ftype == 'd')) {
					tok = line + strlen(line);
					while (*--tok != ' ') {
						if (tok < line)
							goto skip;
					}
					fname[0] = ftype;
					fname[1] = ' ';
					fname[2] = '\0';
					STRNCAT(fname, tok + 1);
					AddLine(lines, fname);
				}
			}
		}
		result = FTPEndDataCmd(cip, 1);
		FTPStopIOTimer(cip);
		if (result < 0) {
			result = kErrLISTFailed;
			cip->errNo = result;
		}
		result = kNoErr;
	} else if (result == kErrGeneric) {
		result = kErrLISTFailed;
		if ((cip->errNo == kErrAcceptDataSocket) || (cip->errNo == kErrNewStreamSocket))
			FTPCloseHost(cip);
		cip->errNo = result;
	}
#endif	/* NO_SIGNALS */

	return (result);
}	/* FileLIST */



static int
FileNLST(const FTPCIPtr cip, LineListPtr lines, const char *lsflags)
{
	char secondaryBuf[512];
	char line[128];
	char *cp1, *cp2, *tok;
	int ftype;
	char fname[128];
#ifndef NO_SIGNALS
	char *secBufPtr, *secBufLimit;
	int nread;
	int result;
#else	/* NO_SIGNALS */
	SReadlineInfo lsSrl;
	int result;
#endif	/* NO_SIGNALS */

	InitLineList(lines);

	FTPInitIOTimer(cip);
	FTPStartIOTimer(cip);
	result = FTPStartDataCmd(cip, kNetReading, kTypeAscii, (longest_int) 0, "%s%s", "NLST", lsflags);

#ifdef NO_SIGNALS

	if (result == 0) {
		if (InitSReadlineInfo(&lsSrl, cip->dataSocket, secondaryBuf, sizeof(secondaryBuf), (int) cip->xferTimeout, 1) < 0) {
			/* Not really fdopen, but close in what we're trying to do. */
			result = kErrFdopenR;
			cip->errNo = kErrFdopenR;
			/* Error(cip, kDoPerror, "Could not fdopen.\n"); */
			return (result);
		}
		
		for (;;) {
			result = SReadline(&lsSrl, line, sizeof(line) - 2);
			if (result == kTimeoutErr) {
				/* timeout */
				/* Error(cip, kDontPerror, "Could not directory listing data -- timed out.\n"); */
				cip->errNo = kErrDataTimedOut;
				return (cip->errNo);
			} else if (result == 0) {
				/* end of listing -- done */
				cip->numListings++;
				break;
			} else if (result < 0) {
				/* error */
				/* Error(cip, kDoPerror, "Could not read directory listing data"); */
				result = kErrLISTFailed;
				cip->errNo = kErrLISTFailed;
				break;
			}

			cip->bytesTransferred += (long) result;
			/* Parse files out of possibly
			 * multicolumn output.
			 */
			for (cp1 = line; ; cp1 = NULL) {
				tok = strtok(cp1, " \r\n\t");
				if (tok == NULL)
					break;
				cp2 = tok + strlen(tok) - 1;
				switch (*cp2) {
				case '*':
				case '@':
					*cp2 = '\0';
					ftype = '-';
					break;
				case '/':
					*cp2 = '\0';
					ftype = 'd';
					break;
				default:
					ftype = '-';
					break;
				}
				fname[0] = (char) ftype;
				fname[1] = (char) ' ';
				fname[2] = '\0';
				STRNCAT(fname, tok);
				AddLine(lines, fname);
			}
		}

		DisposeSReadlineInfo(&lsSrl);
		result = FTPEndDataCmd(cip, 1);
		FTPStopIOTimer(cip);
		if (result < 0) {
			result = kErrLISTFailed;
			cip->errNo = result;
		}
		result = kNoErr;
	} else if (result == kErrGeneric) {
		result = kErrLISTFailed;
		if ((cip->errNo == kErrAcceptDataSocket) || (cip->errNo == kErrNewStreamSocket))
			FTPCloseHost(cip);
		cip->errNo = result;
	}

#else	/* NO_SIGNALS */
	
	if (result == 0) {
		/* This line sets the buffer pointer so that the first thing
		 * BufferGets will do is reset and fill the buffer using
		 * real I/O.
		 */
		secBufPtr = secondaryBuf + sizeof(secondaryBuf);
		secBufLimit = (char *) 0;

		for ( ; ; ) {
			nread = BufferGets(line, sizeof(line), cip->dataSocket, secondaryBuf, &secBufPtr, &secBufLimit, sizeof(secondaryBuf));
			if (nread <= 0) {
				if (nread < 0)
					break;
			} else {
				cip->bytesTransferred += (long) nread;
				/* Parse files out of possibly
				 * multicolumn output.
				 */
				for (cp1 = line; ; cp1 = NULL) {
					tok = strtok(cp1, " \r\n\t");
					if (tok == NULL)
						break;
					cp2 = tok + strlen(tok) - 1;
					switch (*cp2) {
						case '*':
						case '@':
							*cp2 = '\0';
							ftype = '-';
							break;
						case '/':
							*cp2 = '\0';
							ftype = 'd';
							break;
						default:
							ftype = '-';
							break;
					}
					fname[0] = ftype;
					fname[1] = ' ';
					fname[2] = '\0';
					STRNCAT(fname, tok);
					AddLine(lines, fname);
				}
			}
		}
		result = FTPEndDataCmd(cip, 1);
		FTPStopIOTimer(cip);
		if (result < 0) {
			result = kErrLISTFailed;
			cip->errNo = result;
		}
		result = kNoErr;
	} else if (result == kErrGeneric) {
		result = kErrLISTFailed;
		if ((cip->errNo == kErrAcceptDataSocket) || (cip->errNo == kErrNewStreamSocket))
			FTPCloseHost(cip);
		cip->errNo = result;
	}
#endif	/* NO_SIGNALS */

	return (result);
}	/* FileNLST */




static double
Duration(struct timeval *t0)
{
	struct timeval t1;
	double sec;

	gettimeofday(&t1, NULL);
	if (t0->tv_usec > t1.tv_usec) {
		t1.tv_usec += 1000000;
		t1.tv_sec--;
	}
	sec = ((double) (t1.tv_usec - t0->tv_usec) * 0.000001)
		+ (t1.tv_sec - t0->tv_sec);

	return (sec);
}	/* Duration */



static int
FileList(const FTPCIPtr cip)
{
	int p, rc;
	LinePtr lp;
	struct timeval t0;
	double dura;

	if (gHaveDir == 1) {
		DisposeLineListContents(&gDir);
		gHaveDir = 0;
	}
	p = Ri(3);
	if (p == 0) {
		printf("%-5d ls -F", gLoops);
		fflush(stdout);
		gettimeofday(&t0, NULL);
		rc = FileNLST(cip, &gDir, " -F");
	} else if (p == 1) {
		printf("%-5d dir", gLoops);
		fflush(stdout);
		gettimeofday(&t0, NULL);
		rc = FileLIST(cip, &gDir, "");
	} else {
		printf("%-5d ls -CF", gLoops);
		fflush(stdout);
		gettimeofday(&t0, NULL);
		rc = FileNLST(cip, &gDir, " -CF");
	}
	if (rc >= 0) {
		gHaveDir = 1;
		dura = Duration(&t0);
		printf("\t%.3f sec\t%ld bytes %.3f sec %.3f kB/sec\n", dura, (long) cip->bytesTransferred, cip->sec, cip->kBytesPerSec);
		if (gPrintLists != 0) {
			for (lp=gDir.first; lp != NULL; lp=lp->next)
				printf("    %s\n", lp->line);
		}
	}
	return (rc);
}	/* FileList */




static char *
RandomFile(void)
{
	LinePtr lp;
	int nfiles, pick, i;

	if (gHaveDir == 0)
		return NULL;

	for (lp=gDir.first, nfiles=0; lp != NULL; lp=lp->next)
		if (lp->line[0] == '-')
			nfiles++;
	if (nfiles == 0)
		return NULL;

	pick = Ri(nfiles);
	for (lp=gDir.first, i=0; lp != NULL; lp=lp->next) {
		if (lp->line[0] == '-') {
			if (pick == i)
				return (lp->line + 2);
			i++;
		}
	}
	return NULL;
}	/* RandomFile */




static char *
RandomDir(void)
{
	LinePtr lp;
	int ndirs, pick, i;

	if (gHaveDir == 0)
		return NULL;

	for (lp=gDir.first, ndirs=0; lp != NULL; lp=lp->next)
		if (lp->line[0] == 'd')
			ndirs++;
	if (ndirs == 0)
		return NULL;

	pick = Ri(ndirs);
	for (lp=gDir.first, i=0; lp != NULL; lp=lp->next) {
		if (lp->line[0] == 'd') {
			if (pick == i)
				return (lp->line + 2);
			i++;
		}
	}
	return NULL;
}	/* RandomDir */



static void
Shell(const FTPCIPtr cip)
{
	int i, maxw, needlist = 1;
	int p;
	char *randdir, *randfile;
	char s1[128];
	struct timeval t0;
	double dura;
	time_t now;

	for (i=1; i<c_max; i++) {
		gWeights[i] += gWeights[i-1];
	}
	maxw = gWeights[i-1];
	for (gLoops = 0; ((gMaxLoops < 1) || (gLoops < gMaxLoops)); gLoops++) {
		time(&now);
		if ((gAlarmClock != 0) && (now >= gAlarmClock)) {
			printf("%-5d time up!\n", gLoops);
			return;
		}
		p = Ri(maxw);
		for (i=0; i<c_max; i++) {
			if (p < gWeights[i])
				break;
		}

		/* Lost connection? */
		if (cip->connected == 0)
			i = c_quit;

		switch (i) {
			case c_quit:
				if (gQuitMode == kYesCloseYesQuit) {
					printf("%-5d quit\n", gLoops);
					return;
				} else if (gQuitMode == kYesCloseNoQuit) {
					printf("%-5d close", gLoops);
					fflush(stdout);
					gettimeofday(&t0, NULL);
					if ((FTPCloseHost(cip)) < 0) {
						dura = Duration(&t0);
						printf("\t%.3f sec\n", dura);
						fflush(stdout);
						fprintf(kErrStream, "Cannot close host.\n");
						DisposeWinsock();
						exit(5);
					}
					dura = Duration(&t0);
					printf("\t%.3f sec\n", dura);
					sleep((unsigned int) Ri(3) + 3);
					needlist = 1;
					printf("\n%-5d open", ++gLoops);
					fflush(stdout);
					gettimeofday(&t0, NULL);
					STRNCPY(cip->pass, gPasswd);
					if ((FTPOpenHost(cip)) < 0) {
						dura = Duration(&t0);
						printf("\t%.3f sec\n", dura);
						fflush(stdout);
						fprintf(kErrStream, "Cannot open host.\n");
						(void) fprintf(kErrStream, "Cannot open %s: %s.\n", cip->host, FTPStrError(cip->errNo));
						DisposeWinsock();
						exit(5);
					}
					if (gNoMlst != 0) {
						cip->hasMLST = cip->hasMLSD = 0;
					}

					dura = Duration(&t0);
					printf("\t%.3f sec\n", dura);
				} /* else kNoCloseNoQuit */
				break;
			case c_cd:
				if (Rp(20)) {
					printf("%-5d cd ..", gLoops);
					fflush(stdout);
					gettimeofday(&t0, NULL);
					(void) FTPChdir(cip, "..");
					dura = Duration(&t0);
					printf("\t%.3f sec\n", dura);
					needlist = 1;
				} else if (Rp(20)) {
					printf("%-5d cd %s", gLoops, gMainDir);
					fflush(stdout);
					gettimeofday(&t0, NULL);
					(void) FTPChdir(cip, gMainDir);
					dura = Duration(&t0);
					printf("\t%.3f sec\n", dura);
					needlist = 1;
				} else {
cd:
					if (needlist) {
						if (FileList(cip) >= 0)
							needlist = 0;
						++gLoops;
					}
					if ((needlist == 0) && ((randdir = RandomDir()) != NULL)) {
						printf("%-5d cd %s", gLoops, randdir);
						fflush(stdout);
						gettimeofday(&t0, NULL);
						if (FTPChdir(cip, randdir) == 0)
							needlist = 1;
						dura = Duration(&t0);
						printf("\t%.3f sec\n", dura);
					} else {
						printf("%-5d cd %s", gLoops, gMainDir);
						fflush(stdout);
						gettimeofday(&t0, NULL);
						(void) FTPChdir(cip, gMainDir);
						dura = Duration(&t0);
						printf("\t%.3f sec\n", dura);
						needlist = 1;
					}
				}
				break;
			case c_pwd:
				printf("%-5d pwd", gLoops);
				fflush(stdout);
				gettimeofday(&t0, NULL);
				(void) FTPGetCWD(cip, s1, sizeof(s1));	/* pwd */
				dura = Duration(&t0);
				printf("\t%.3f sec\n", dura);
				break;
			case c_dir:
				if (FileList(cip) >= 0)
					needlist = 0;
				break;
			case c_get:
				if (needlist) {
					if (FileList(cip) >= 0)
						needlist = 0;
					++gLoops;
				}
				if (needlist == 0) {
					randfile = RandomFile();
					if (randfile == NULL)
						goto cd;
					/* pick a random file to get */
					printf("%-5d get %s", gLoops, randfile);
					fflush(stdout);
					gettimeofday(&t0, NULL);
					(void) FTPReadOneFile(cip, randfile);
					dura = Duration(&t0);
					printf("\t%.3f sec\t%ld bytes %.3f sec %.3f kB/sec\n", dura, (long) cip->bytesTransferred, cip->sec, cip->kBytesPerSec);
				}
				break;
		}
		sleep((unsigned int) Ri(4));
	}
}	/* Shell */



main_void_return_t
main(int argc, char **argv)
{
	int result, c;
	int seed;
	time_t now;
	struct tm *ltp;
	char dstr[64];
	char *password;
	GetoptInfo opt;

	InitWinsock();

	result = FTPInitLibrary(&li);
	if (result < 0) {
		fprintf(kErrStream, "Init library error %d.\n", result);
		DisposeWinsock();
		exit(3);
	}
	result = FTPInitConnectionInfo(&li, &fi, kDefaultFTPBufSize);
	if (result < 0) {
		fprintf(kErrStream, "Init connection info error %d.\n", result);
		DisposeWinsock();
		exit(4);
	}

	fi.debugLog = NULL;
	fi.errLog = kErrStream;
	STRNCPY(fi.user, "anonymous");
	STRNCPY(fi.pass, "monkey@bowser.nintendo.co.jp");
	seed = -1;

	GetoptReset(&opt);
	while ((c = Getopt(&opt, argc, argv, "a:QRSD:m:ls:d:e:P:p:u:")) > 0) switch(c) {
		case 'a':
			time(&gAlarmClock);
			gAlarmClock += (time_t) atoi(opt.arg);
			break;
		case 'm':
			gMaxLoops = atoi(opt.arg);
			break;
		case 's':
			seed = atoi(opt.arg);
			break;
		case 'D':
			gMainDir = opt.arg;
			break;
		case 'l':
			gPrintLists = !gPrintLists;
			break;
		case 'Q':
			gQuitMode = kNoCloseNoQuit;
			break;
		case 'R':
			gQuitMode = kYesCloseNoQuit;
			break;
		case 'S':
			gQuitMode = kYesCloseYesQuit;
			break;
		case 'P':
			fi.port = atoi(opt.arg);	
			break;
		case 'u':
			STRNCPY(fi.user, opt.arg);
			break;
		case 'p':
			STRNCPY(fi.pass, opt.arg);	/* Don't recommend doing this! */
			break;
		case 'e':
			fi.errLog = fopen(opt.arg, "a");
			break;
		case 'd':
			fi.debugLog = fopen(opt.arg, "a");
			break;
		default:
			Usage();
	}
	if (opt.ind > argc - 1)
		Usage();

	if (strcmp(fi.user, "anonymous") && strcmp(fi.user, "ftp")) {
		if (fi.pass[0] == '\0') {
			password = GetPass2("Password: ");		
			if (password != NULL) {
				STRNCPY(fi.pass, password);
				/* Don't leave cleartext password in memory. */
				memset(password, 0, strlen(fi.pass));
			}
		}
	}

	/* The library destroys it each time. */
	STRNCPY(gPasswd, fi.pass);

	if (seed < 0) {
		seed = time(NULL) & 0x7fff;
		fprintf(stdout, "Seed is %d.\n", seed);
	}
	srand((unsigned int) seed);

	STRNCPY(fi.host, argv[opt.ind]);
	
	if ((result = FTPOpenHost(&fi)) < 0) {
		(void) fprintf(stderr, "Cannot open %s: %s.\n", fi.host, FTPStrError(result));
		DisposeWinsock();
		exit(5);
	}
	if (gNoMlst != 0) {
		fi.hasMLST = fi.hasMLSD = 0;
	}
	time(&now);
	ltp = localtime(&now);
	strftime(dstr, sizeof(dstr), "%Y-%m-%d %H:%M:%S", ltp);
	printf("-- Started at %s --\n", dstr);
	Shell(&fi);
	FTPCloseHost(&fi);
	

	time(&now);
	ltp = localtime(&now);
	strftime(dstr, sizeof(dstr), "%Y-%m-%d %H:%M:%S", ltp);
	printf("-- Finished at %s --\n", dstr);
	
	DisposeWinsock();
	exit(0);
}	/* main */
